!==============================================================================!
 module arnoldi_mod
!==============================================================================!
    implicit none
    private
!
    integer, parameter :: sp = kind(1.0)
    integer, parameter :: dp = kind(1.0d0)
    integer, parameter :: zp = kind((0.0d0,0.0d0))
!
    complex(zp), parameter :: one = complex(1.0d0, 0.0d0)
    complex(zp), parameter :: zero = complex(0.0d0, 0.0d0)
!
! Public procedures
!
    public :: arnoldi_iteration 
!
! Private data
!
! ARPACK's debug common block
!
    integer :: logfil = 6, ndigit = -3, mcaupd = 1
    integer :: mgetv0 = 0, msaupd = 0, msaup2 = 0, msaitr = 0, mseigt = 0,  &
               msapps = 0, msgets = 0, mseupd = 0, mnaupd = 0, mnaup2 = 0,  &
               mnaitr = 0, mneigh = 0, mnapps = 0, mngets = 0, mneupd = 0,  &
               mcaup2 = 0, mcaitr = 0, mceigh = 0, mcapps = 0, mcgets = 0,  &
               mceupd = 0
!
    common /debug/ logfil, ndigit, mgetv0, msaupd, msaup2, msaitr, mseigt,  &
                   msapps, msgets, mseupd, mnaupd, mnaup2, mnaitr, mneigh,  &
                   mnapps, mngets, mneupd, mcaupd, mcaup2, mcaitr, mceigh,  &
                   mcapps, mcgets, mceupd
!==============================================================================!
 contains
!==============================================================================!
! Public procedures
!==============================================================================!
    subroutine arnoldi_iteration (ndim, a, nev, d)
!
! Use ARPACK routines to find a few eigenvalues (lambda) and corresponding 
! eigenvectors (x) for the standard eigenvalue problem:
!          
!                        A*x = lambda*x
! 
! where A is a general NDIM by NDIM complex matrix
!
! This subroutine is based on an ARPACK test program adapted by Logan Boudet
! as part of his M1 project in Rennes in 2022.
! 
    integer,     intent(in)    :: ndim
    complex(zp), intent(in)    :: a(:,:)
    integer,     intent(inout) :: nev           ! number of eigenvalues required
    complex(zp), intent(out)   :: d(:)          ! vector of required eigenvalues
!
! Local data
!
    character(1), parameter :: bmat = "I"       ! standarg eigenvalue problem
    character(2), parameter :: which = "LM"     ! find NEV eigenvalues of 
!                                               ! largest magnitude
!
    integer     :: ido, ierr, info, j, lworkl, nconv, ncv 
    integer     :: iparam(11)
    integer     :: ipntr(14)
    real(dp)    :: tol
    complex(zp) :: sigma
    logical     :: rvec
!
    real(dp),    allocatable :: rd(:,:)
    real(dp),    allocatable :: rwork(:)
    complex(zp), allocatable :: ax(:)
    complex(zp), allocatable :: resid(:)
    complex(zp), allocatable :: v(:,:)
    complex(zp), allocatable :: workd(:)
    complex(zp), allocatable :: workev(:)
    complex(zp), allocatable :: workl(:)
    logical,     allocatable :: select(:)
!
! External BLAS/LAPACK functions used
!
    real(dp), external :: dznrm2, dlapy2 
!
!
    write(6,*) "----------------- BEGIN OF ARNOLDI ITERATION -----------------"
!
! NCV is is the largest number of basis vectors that will be used in the 
! Implicitly Restarted Arnoldi Process. Work per major iteration is 
! proportional to NDIM*NCV*NCV.
! 
! Note: we must have NCV >= NEV + 2, and preferably NCV >= 2*NEV
!
!    ncv = max(ceiling(1.125*nev + 15), nev+3)
    ncv = 2*nev
!
    iparam(11) = 0
    ipntr(14) = 0
!
! stopping criteria; machine precision is used if tol <= 0
!
    tol = 0.0_dp
!
! Algorithm mode
!
    iparam(1) = 1           ! exact shift strategy              
    iparam(3) = 300         ! maximum number of iterations
    iparam(7) = 1           ! use mode1 of ZNAUPD
!
    lworkl = ncv*(3*ncv + 5)
    allocate(rwork(ncv))
    allocate(resid(ndim), v(ndim,ncv))
    allocate(workd(3*ndim), workev(2*ncv), workl(lworkl))
    allocate(select(ncv))
!
! IDO is the reverse communication parameter used to determine action to be taken
! on return from ZNAUPD. Its initial value must be 0.
!
    ido = 0
!
! On entry, INFO == 0 instructs ZNAUPD to use a random starting vector.
! To specify a particular starting vector, set INFO to a non-zero value.
! The required startng vector should then be supplied in RESID.
!    
    info = 0
!
! Main reverse communication loop
!
    do
        call znaupd(ido, bmat, ndim, which, nev, tol, resid, ncv, v, ndim,  &
                    iparam, ipntr, workd, workl, lworkl, rwork, info)

        if (abs(ido) /= 1) exit
!
! Matrix-vector multiplication y = A*x
! Initial vector x is stored starting at  workd(ipntr(1))
! The result y should be stored in workd(ipntr(2))
!
        call matvec(a, ndim, ndim, workd(ipntr(1)), workd(ipntr(2)))
!
    end do
!
    if (info < 0) then
        write(6,*)
        write(6,*) "Error: znaupd returned info = ", info
        write(6,*) "Check the documentation of znaupd for more information"
        write(6,*)
        stop
    end if
!
! No fatal errors, post-process with ZNEUPD to extract computed eigenvalues.
! Eigenvectors may be also computed by setting RVEC = .TRUE.)
!
    if (info == 1) then
        write(6,*)
        write(6,*) "Maximum number of iterations reached."
        write(6,*)
    else if (info == 3) then
        write(6,*)
        write(6,*) "No shifts could be applied during implicit Arnoldi update"
        write(6,*) "Try increasing NCV."
        write(6,*)
    end if      
!
    rvec = .false.
!
    call zneupd(rvec, 'A', select, d, v, ndim, sigma, workev, bmat, ndim,    &
                which, nev, tol, resid, ncv, v, ndim, iparam, ipntr, workd,  &
                workl, lworkl, rwork, ierr)
!
    if (ierr /= 0) then
        write(6,*)
        write(6,*) "Error: zneupd returned ierr = ", ierr
        write(6,*) "Check the documentation of zneupd for more information"
        write(6,*)
        stop
    end if
!
! Eigenvalues are returned in the one dimensional array D and if RVEC == .TRUE.
! the corresponding eigenvectors are returned in the first NCONV == IPARAM(5) 
! columns of the two dimensional array V 
! 
    nconv = iparam(5)
!
    if (rvec) then
!
! Compute the residual norm ||  A*x - lambda*x || for the NCONV accurately 
! computed eigenvalues and eigenvectors
!
        allocate(rd(ncv,3), ax(ndim))
!
        do j = 1, nconv
            call matvec(a, ndim, ndim, v(:,j), ax)
            call zaxpy(ndim, -d(j), v(:,j), 1, ax, 1)
            rd(j,1) = real(d(j), dp)
            rd(j,2) = aimag(d(j))
            rd(j,3) = dznrm2(ndim, ax, 1) / dlapy2(rd(j,1), rd(j,2))
        end do
!
        call dmout(6, nconv, 3, rd, 2*nev, -6,  &
                   "Ritz values (Real, Imag) and relative residuals")
!
        deallocate(rd, ax)
!
    end if
!
    write(6,*)
    write(6,*) "SUMMARY"
    write(6,*) "======="
    write(6,*)
    write(6,*) "Size of the matrix is ", ndim
    write(6,*) "The number of Ritz values requested is ", nev
    write(6,*) "The number of Arnoldi vectors generated (NCV) is ", ncv
    write(6,*) "Portion of the spectrum: ", which
    write(6,*) "The number of converged Ritz values is ", nconv 
    write(6,*) "The number of implicit Arnoldi update iterations taken is ", iparam(3)
    write(6,*) "The number of OP*x is ", iparam(9)
    write(6,*) "The convergence criterion is ", tol
    write(6,*) 
!
    nev = nconv
!   
    write(6,*) "------------------ END OF ARNOLDI ITERATION ------------------"
!
    deallocate(rwork)
    deallocate(resid, v)
    deallocate(workd, workev, workl)
    deallocate(select)
!
    return
    end subroutine arnoldi_iteration
!==============================================================================!
! Private procedures
!==============================================================================!
    subroutine matvec (a, n, lda, x, y)
!
! Compute the matrix-vector product a*x, storing the result in y
!
    complex(zp), intent(in)  :: a(:,:)
    integer,     intent(in)  :: n
    integer,     intent(in)  :: lda
    complex(zp), intent(in)  :: x(*)
    complex(zp), intent(out) :: y(*)
!
!
    call zgemv('n', n, n, one, a, lda, x, 1, zero, y, 1)
!
    return
    end subroutine matvec
!==============================================================================!
 end module arnoldi_mod
!==============================================================================!
